#Copyright 2015 Yale University - Grablab
#Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

#!/usr/bin/env python
from ycb_benchmarks import * 
from ycb_benchmarks.srv import * 
import roslib
import os
import sys
import json
import urllib
import urllib2
import ycb
import rospy
import rospkg
import shutil
import h5py as h5
import numpy as np
from scipy.misc import imread
import IPython
import math

base_url = "http://ycb-benchmarks.s3-website-us-east-1.amazonaws.com/data/"
objects_url = base_url + "objects.json"
try:
	data_base_directory = rospy.get_param("/ycb_benchmarks/data_base_directory")   
	print "[ycb_benchmarks] Output directory is set to %s." % data_base_directory
except KeyError, getparam_error:
	rospack = rospkg.RosPack()
	package_path = rospack.get_path('ycb_benchmarks');
	data_base_directory = package_path + "/data"
	print "[ycb_benchmarks] The parameter %s is not set in the parameter server. It is set to %s by default." % (str(getparam_error),data_base_directory)

if os.access(data_base_directory, os.F_OK) == False:
	try:
		os.mkdir(data_base_directory)
	except OSError, mkdir_error:
		print "[ycb_benchmarks] Failed to create folder %s. Error message: %s" % (data_base_directory, str(mkdir_error))
		exit(1)
if os.access(data_base_directory, os.W_OK) == False:
	print "[ycb_benchmarks] The output directory %s is not writable. Aborting." % data_base_directory
	exit(1)

# Extract all files from the downloaded .tgz, and remove .tgz files.
# If false, will just download all .tgz files to data_base_directory
extract = True

if not os.path.exists(data_base_directory):
    os.makedirs(data_base_directory)

def handle_generate_point_cloud(req):
	objects_to_pc = [] 
	views_to_pc = []
	view_angles_to_pc = []
	ids_to_pc = req.object_id
	if len(ids_to_pc) == 0:
		res = GeneratePointCloudResponse()
		res.success = False
		error_message = "Object id array is empty."
		return res
	views_to_pc = req.viewpoint_camera
	if len(views_to_pc) == 0:
		res = GeneratePointCloudResponse()
		res.success = False
		res.error_message = "Point Cloud viewpoint array is empty."
		return res
	view_angles_to_pc = req.viewpoint_angle
	if len(view_angles_to_pc) == 0:
		res = GeneratePointCloudResponse()
		res.success = False
		res.error_message = "Reference camera array is empty."
		return res
	try:
		file_object_ids = open(package_path + "/object_ids.txt", 'r')
	except IOError as error_open_file:
		print "[ycb_benchmarks] Error reading the object_ids.txt file in ycb_benchmarks package. Aborting."
		exit(1)

	for id_str in ids_to_pc:
		if len(id_str) == 0:
			res = GeneratePointCloudResponse()
			res.success = False
			res.error_message = "Object id array has empty string."
			return res;
		elif id_str.find('-') != -1:
			while len(id_str) < 5:
				id_str = "0"+id_str
		else:
			while len(id_str) < 3:
				id_str = "0"+id_str
		for file_line in file_object_ids:
			if id_str in file_line: 
#				file_line = file_line[:-1]
				word = file_line.split(None, 1)[0]
				objects_to_pc.append(word)
	file_object_ids.close()
	ref = 5
	for object_str in objects_to_pc:
		for view in views_to_pc:
			if view <= 0 or view >5:
				res = GeneratePointCloudResponse()
				res.success = False
				res.error_message = "[ycb_benchmarks] Viewpoint camera should be an integer between 1 and 5. The input was %d." % angle
				return res
			for angle in view_angles_to_pc:
				if angle%3 != 0 or angle<0 or angle >357:
					res = GeneratePointCloudResponse()
					res.success = False
					res.error_message =  "[ycb_benchmarks] Angle should be a multiple of 3, and between 0 and 357. The input was %d. Skipping this point cloud generation request." % angle
					return res
				basename = "NP{0}_{1}".format(str(view), str(angle))
				depthFilename = os.path.join(data_base_directory +"/" + object_str, basename + ".h5")
				rgbFilename = os.path.join(data_base_directory +"/"+ object_str, basename + ".jpg")
				calibrationFilename = os.path.join(data_base_directory+"/"+object_str, "calibration.h5")
				calibration = h5.File(calibrationFilename)
				rgbImage = imread(rgbFilename)
				depthK = calibration["NP{0}_ir_K".format(str(view))][:]
				rgbK = calibration["NP{0}_rgb_K".format(str(view))][:]
				depthScale = np.array(calibration["NP{0}_ir_depth_scale".format(str(view))]) * .0001 # 100um to meters
				H_RGBFromDepth = ycb.getRGBFromDepthTransform(calibration, view, ref)
				unregisteredDepthMap = h5.File(depthFilename)["depth"][:]
				unregisteredDepthMap = ycb.filterDiscontinuities(unregisteredDepthMap) * depthScale

				registeredDepthMap = ycb.registerDepthMap(unregisteredDepthMap,
                                         rgbImage,
                                         depthK,
                                         rgbK,
                                         H_RGBFromDepth)

				pointCloud = ycb.registeredDepthMapToPointCloud(registeredDepthMap, rgbImage, rgbK)

				if os.access(data_base_directory+"/"+object_str+"/clouds", os.F_OK) == False:
					try:
						os.mkdir(data_base_directory+"/"+object_str+"/clouds")
					except OSError, mkdir_error:
						print "[ycb_benchmarks] Failed to create folder %s. Error message: %s" % (data_base_directory, str(mkdir_error))
						exit(1)

				ycb.writePLY(data_base_directory+"/"+object_str+"/clouds/pc_NP"+str(view)+"_NP"+str(ref)+"_"+str(angle)+".ply", pointCloud)

	res = GeneratePointCloudResponse()
	res.success = True
	res.error_message = ""
	return res


def handle_delete_data(req):

	objects_to_delete =[] 
	ids_to_delete = req.object_id
	if len(ids_to_delete) == 0:
		res = DataDeleteResponse()
		res.success = False
		error_message = "Object id array is empty."
		return res
	try:
		file_object_ids = open(package_path + "/object_ids.txt", 'r')
	except IOError as error_open_file:
		print "[ycb_benchmarks] Error reading the object_ids.txt file in ycb_benchmarks package. Aborting."
		exit(1)

	for id_str in ids_to_delete:
		if len(id_str) == 0:
			res = DataDeleteResponse()
			res.success = False
			res.error_message = "Object id array has empty string."
			return res;
		elif id_str.find('-') != -1:
			while len(id_str) < 5:
				id_str = "0"+id_str
		else:
			while len(id_str) < 3:
				id_str = "0"+id_str
		for file_line in file_object_ids:
			if id_str in file_line: 
				word = file_line.split(None, 1)[0]
				objects_to_delete.append(word)
	file_object_ids.close()
	for object_str in objects_to_delete:
		filename = "{path}/{object}".format(path=data_base_directory,
                                                                  object=object_str)
		try:
			shutil.rmtree(filename)
		except (IOError, OSError) as error_delete:
			res = DataDeleteResponse()
			res.success = False
			res.error_message = "Data cannot be deleted. Reason: %s" %str(error_delete)
			return res
	res = DataDeleteResponse()
	res.success = True
	res.error_message = ""
	return res

def handle_download_data(req):
	objects_to_download =[] 
	files_to_download = []

	ids_to_download = req.object_id
	if len(ids_to_download) == 0:
		res = DataCommandResponse()
		res.success = False
		error_message = "Object id array is empty."
		return res
	files_to_download = req.data_type
	if len(files_to_download) == 0 or (len(files_to_download)==1 and files_to_download[0] == ''):
		res = DataCommandResponse()
		res.success = False
		res.error_message = "File type array is empty."
		return res

	try:
		file_object_ids = open(package_path + "/object_ids.txt", 'r')
	except IOError as error_open_file:
		print "error reading the object_ids.txt file in ycb_benchmarks package. Aborting."
		res = DataCommandResponse()
		res.success = False
		res.error_message = "object_ids.txt is not found in the ycb_benchmarks package folder. Aborting."
		return res

	for id_str in ids_to_download:
		if len(id_str) == 0:
			res = DataCommandResponse()
			res.success = False
			res.error_message = "Object id array has empty string."
			return res;
		elif id_str.find('-') != -1:
			while len(id_str) < 5:
				id_str = "0"+id_str
		else:
			while len(id_str) < 3:
				id_str = "0"+id_str

		for file_line in file_object_ids:
			if id_str in file_line: 
#				objects_to_download.append(file_line.split()[0])
#				print "\n hello %s what\n" % file_line.split()[0]

#				file_line = file_line[:-1]
				word = file_line.split()[0]
				objects_to_download.append(word)
	file_object_ids.close()
	for object_str in objects_to_download:
		for file_type in files_to_download:
			url = ycb.tgz_url(object_str, file_type, base_url)
			print url
			if not ycb.check_url(url):
				res = DataCommandResponse()
				res.success = False
				res.error_message = "The %s data of the object %s cannot be found in the server." % (file_type, object_str)
				return res
			filename = "{path}/{object}_{file_type}.tgz".format(path=data_base_directory,
                                                                    object=object_str,
                                                                    file_type=file_type)


			ycb.download_file(url, filename)

			if extract:
				if file_type in ["berkeley_rgbd", "berkeley_rgb_highres", "berkeley_meshes"]:
					ycb.extract_tgz(filename, data_base_directory)
				else:
					ycb.extract_tgz(filename, data_base_directory)

	res = DataCommandResponse()
	res.success = True
	res.error_message = ""
	return res
	
def handle_generate_urdf(req):
	objects_to_urdf =[] 
	files_to_urdf = []
	object_mass = [] 

	ids_to_urdf = req.object_id
	if len(ids_to_urdf) == 0:
		res = DataCommandResponse()
		res.success = False
		error_message = "Object id array is empty."
		return res
	files_to_urdf = req.data_type
	if len(files_to_urdf) == 0 or (len(files_to_urdf)==1 and files_to_urdf[0] == ''):
		res = DataCommandResponse()
		res.success = False
		res.error_message = "File type array is empty."
		return res

	try:
		file_object_ids = open(package_path + "/object_ids.txt", 'r')
	except IOError as error_open_file:
		print "error reading the object_ids.txt file in ycb_benchmarks package. Aborting."
		res = DataCommandResponse()
		res.success = False
		res.error_message = "object_ids.txt is not found in the ycb_benchmarks package folder. Aborting."
		return res

	for id_str in ids_to_urdf:
		if len(id_str) == 0:
			res = DataCommandResponse()
			res.success = False
			res.error_message = "Object id array has empty string."
			return res;
		elif id_str.find('-') != -1:
			while len(id_str) < 5:
				id_str = "0"+id_str
		else:
			while len(id_str) < 3:
				id_str = "0"+id_str

		for file_line in file_object_ids:

			if id_str in file_line: 
				word = file_line.split(None, 1)[0]
				objects_to_urdf.append(word)
				word = file_line.split(None, 1)[1]
				object_mass.append(word[:-1])
				break
	file_object_ids.close()
	it = 0
	for object_str in objects_to_urdf:
		for file_type in files_to_urdf:
			ycb.generate_urdf(object_str,file_type,object_mass[it],data_base_directory)
			it = it+1

	res = DataCommandResponse()
	res.success = True
	res.error_message = ""
	return res

if __name__ == "__main__":
	rospy.init_node('ycb_benchmarks')
	srv_download_data = rospy.Service('/ycb_benchmarks/download_data',DataCommand,handle_download_data)
	srv_delete_data = rospy.Service('/ycb_benchmarks/delete_data',DataDelete,handle_delete_data)
	srv_generate_point_cloud = rospy.Service('/ycb_benchmarks/generate_point_cloud',GeneratePointCloud,handle_generate_point_cloud)
	srv_create_urdf = rospy.Service('/ycb_benchmarks/generate_urdf',DataCommand,handle_generate_urdf)
	rospy.spin()
	exit(0)
